/*
** nanoasm compiler (C) 2007, Alexander G. Karpov a.k.a. mrhx.
** version 1.0
** January 7, 2007.
**
** http://mrhx.narod.ru/  [ru]
** http://mrhx.t35.com/   [en]
*/
#include <windows.h>
#include "../nanoasm.c"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nanoasm_win32_exe/array.c" // include win32 exe module.

//#pragma comment(linker,"/NODEFAULTLIB")
//#pragma comment(linker,"/ENTRY:main")
//#pragma comment(linker,"/FILEALIGN:512")
//#pragma comment(linker,"/MERGE:.data=.text")
//#pragma comment(linker,"/MERGE:.rdata=.text")
//#pragma comment(linker,"/SECTION:.text,ERWX")
//#pragma comment(linker,"/IGNORE:4078")

void * _stdcall GetAddressOfProc(const char *LibName, const char *ProcName)
{
	HMODULE hMod;
	void *Proc;

	printf("GetAddressOfProc: lib \"%s\" proc \"%s\"\n", LibName, ProcName);

	hMod = LoadLibrary(LibName);
	if (!hMod)
	{
		printf("GetAddressOfProc: could not load library \"%s\"\n", LibName);
		return NULL;
	}

	Proc = GetProcAddress(hMod, ProcName);
	if (!Proc) printf("GetAddressOfProc: could not get address of procedure \"%s\"\n", ProcName);
	else printf("GetAddressOfProc: address is 0x%08X\n", Proc);
	return Proc;
}

int main(int argc, char **args)
{
	FILE *file;
	char *buf, *bin, *pos;
	long size;
	int len, ret;

	printf("nanoasm compiler for win32\nversion 1.0\n(C) 2007, Alexander G. Karpov a.k.a. mrhx\nvisit author`s website: http://mrhx.narod.ru/\n\n");
	if (argc < 2 || argc > 4)
	{
		printf("usage: nanoasm <input file> <mode> [<output file>]\nmode= exe|bin|array|run\n");
		printf("exe- create an executable PE win32 file.\nbin- create a binary file with compiled program.\n");
		printf("array- create a text file with array of bytes to include it into your program.\nrun- just run program without save.\n");
		printf("example1: nanoasm prog.nasm exe prog.exe\n");
		printf("example2: nanoasm prog.nasm run\n");
		return 0;
	}

	printf("open the file \"%s\"...\n", args[1]);
	file = fopen(args[1], "rb");
	if (!file)
	{
		printf("could not open the file\n");
		return 1;
	}
	else printf("ok\n");

	fseek(file, 0, SEEK_END);
	printf("allocate memory %i bytes...\n", (size = ftell(file)) + 1);
	fseek(file, 0, SEEK_SET);
	buf = malloc(size+1);
	if (!buf)
	{
		printf("could not allocate the memory for the file\n");
		fclose(file);
		return 2;
	}
	else printf("ok\n");

	printf("reading the file %i bytes...\n", size);
	if (1 != fread(buf, size, 1, file))
	{
		printf("could not read the file or zero length\n");
		fclose(file);
		free(buf);
		return 3;
	}
	else printf("ok\n");

	fclose(file);
	buf[size] = '\0';

	printf("check the length of output data...\n");
	len = nanoasm_compiled_length(buf);
	printf("output length is %i\n", len);

	printf("allocate memory for output buffer %i bytes...\n", len);
	bin = malloc(len);
	if (!buf)
	{
		printf("could not allocare the memory for output data\n");
		free(buf);
		return 4;
	}
	else printf("ok\n");

	printf("compile the program...\n");
	ret = nanoasm_compile(buf, bin, &pos);
	if (ret < 0)
	{
		printf("error code %i (%s)\n", ret, (ret == -1)?"unknown instruction":"unknown label");
		printf("<here> %.*s\n", 32, pos);
		free(buf);
		free(bin);
		return 5;
	}
	else printf("ok\n");

	free(buf);

	printf("mode: %s...\n", args[2]);
	if (!strcmp(args[2], "bin"))
	{
		printf("create or open output file \"%s\"...\n", args[3]);
		file = fopen(args[3], "wb");
		if (!file)
		{
			printf("could not create or open output file\n");
			free(bin);
			return 7;
		}
		else printf("ok\n");

		printf("write compiled data %i bytes...\n", len);
		if (1 != fwrite(bin, len, 1, file))
		{
			printf("could not write to output file or zero length\n");
			free(bin);
			fclose(file);
			return 8;
		}
		else printf("ok\n");
		fclose(file);
	}
	else if (!strcmp(args[2], "array"))
	{
		int i;

		printf("create or open output file \"%s\"...\n", args[3]);
		file = fopen(args[3], "wt");
		if (!file)
		{
			printf("could not create or open output file\n");
			free(bin);
			return 7;
		}
		else printf("ok\n");

		printf("write compiled data %i bytes...\n", len);

		for (i = 0; i < len; ++i)
		{
			fprintf(file, (i == len-1)?"%3i\n":"%3i, ", (unsigned char) bin[i]);
			if ((i+1)%16 == 0) fprintf(file, "\n");
		}

		printf("ok\n");
		fclose(file);
	}
	else if (!strcmp(args[2], "exe"))
	{
		PIMAGE_NT_HEADERS nthdr;
		PIMAGE_SECTION_HEADER shdr;
		int i, sz, zero = 0;

		printf("create or open output file \"%s\"...\n", args[3]);
		file = fopen(args[3], "wb");
		if (!file)
		{
			printf("could not create or open output file\n");
			free(bin);
			return 7;
		}
		else printf("ok\n");

		// modify exe headers:
		nthdr = (PIMAGE_NT_HEADERS)(array + ((PIMAGE_DOS_HEADER) array)->e_lfanew);
		printf("old size: image %i code %i\n", nthdr->OptionalHeader.SizeOfImage, nthdr->OptionalHeader.SizeOfCode);

		// find procedure to replace:
		for (i = 0; i < sizeof(array); ++i)
			if (*(long*)(array+i) == 0x90C3C033) break;

		if (i == sizeof(array))
		{
			printf("internal data error #1\n");
			free(bin);
			fclose(file);
			return 10;
		}

		printf("patching exe module...\n");

		// calculate rounded size of compiled data:
		sz = nthdr->OptionalHeader.FileAlignment * ((len / nthdr->OptionalHeader.FileAlignment) +
			((len % nthdr->OptionalHeader.FileAlignment)?1:0));

		// insert a jump to compiled proedure:
		array[i] = 0xE9; // jmp i32.
		*(long*)(array+i+1) = nthdr->OptionalHeader.SizeOfCode + nthdr->OptionalHeader.SizeOfHeaders - i - 5;

		nthdr->OptionalHeader.SizeOfCode += sz;

		nthdr->OptionalHeader.SizeOfImage = nthdr->OptionalHeader.SectionAlignment +
			nthdr->OptionalHeader.SectionAlignment*(nthdr->OptionalHeader.SizeOfCode / nthdr->OptionalHeader.SectionAlignment);
		if (nthdr->OptionalHeader.SizeOfCode % nthdr->OptionalHeader.SectionAlignment)
			nthdr->OptionalHeader.SizeOfImage += nthdr->OptionalHeader.SectionAlignment;

		// patching the section .text:
		shdr = (PIMAGE_SECTION_HEADER)((unsigned char*)nthdr + sizeof(IMAGE_NT_HEADERS));
		shdr->SizeOfRawData += sz;
		shdr->Misc.VirtualSize = nthdr->OptionalHeader.SizeOfCode;

		printf("new size: image %i code %i\n", nthdr->OptionalHeader.SizeOfImage, nthdr->OptionalHeader.SizeOfCode);

		printf("write win32_exe module data %i bytes...\n", sizeof(array));
		if (1 != fwrite(array, sizeof(array), 1, file))
		{
			printf("could not write to output file\n");
			free(bin);
			fclose(file);
			return 9;
		}
		else printf("ok\n");

		printf("write compiled data %i bytes...\n", len);
		if (1 != fwrite(bin, len, 1, file))
		{
			printf("could not write to output file or zero length\n");
			free(bin);
			fclose(file);
			return 8;
		}
		else printf("ok\n");

		printf("write block of nulls for alignment %i bytes...\n", i = sz-len);
		for (; i > 0; --i)
			if (1 != fwrite(&zero, 1, 1, file))
			{
				printf("could not write to output file\n");
				free(bin);
				fclose(file);
				return 11;
			}

		fclose(file);
	}
	else if (!strcmp(args[2], "run"))
	{
		int (_cdecl*Proc)(void *) = (int (_cdecl*)(void*)) bin;
		printf("run the program at 0x%08X...\n", Proc);
		printf("the program returns 0x%08X\n", Proc(GetAddressOfProc));
	}
	else
	{
		printf("unknown mode\n");
		free(bin);
		return 6;
	}

	free(bin);
	printf("done\n");
	return 0;
}
